import * as path from "path";
import * as fs from "fs-extra";
import * as klaw from "klaw";
import * as azureStorage from "azure-storage";
import * as https from "https";
import { setTimeout } from "timers";
const svgstore = require("svgstore");


const BASE_URL = "https://skopinion.com/opinionapi/v1";

const TEMPLATES_PATH = path.join(__dirname, "..", "templates");

function getAliasFileName(gamePath: string, instance?: string) {
    if (instance) {
        return path.join(gamePath, `alias-${instance}.json`);
    } else {
        return path.join(gamePath, "alias.json");
    }
}
function getInstanceFileName(gamePath: string, instance: string) {
    return path.join(gamePath, `${instance}.json`);
}

function getIndexFileName(gamePath: string) {
    return path.join(gamePath, "src", "index.json");
}

function getLaunchFileName(gamePath: string) {
    return path.join(gamePath, ".vscode", "launch.json");
}

function getObjectAsync<T>(url: string) {
    return getStringAsync(url).then(s => JSON.parse(s) as T);
}

function getStringAsync(url: string) {
    return new Promise<string>((resolve, reject) => {
        https.get(`${BASE_URL}/${url}`, res => {
            if (res.statusCode !== 200) {
                const error = `Error: ${res.statusCode} (${res.statusMessage})`;
                console.log(error, res.headers);
                reject(error);
            } else {
                let body = "";
                res
                    .on("data", chunk => body += chunk)
                    .on("end", () => {
                        resolve(body);
                    });
            }
        }).on("error", e => reject(e));
    });
}

function bustConfigurationCacheAsync(id: string) {
    process.stdout.write("Breaking cache...");
    const url = `content/configuration/${id}?t=${Date.now()}`;
    return getStringAsync(url).then(() => {
        process.stdout.write("Done\n");
    });
}

function bustAliasCacheAsync(gamePath: string, instance?: string) {
    const aliasFile = getAliasFileName(gamePath, instance);
    fs.exists(aliasFile, exists => {
        if (exists) {
            fs.readJSON(aliasFile, (err, alias) => {
                if (err) {
                    console.log("Failed to read alias", err);
                } else {
                    process.stdout.write(`Breaking alias cache for ${alias.alias}..`);
                    const url = `content/byalias/${alias.alias}?t=${Date.now()}`;
                    getStringAsync(url).then(() => {
                        process.stdout.write(`Breaking game cache for ${alias.alias}..`);
                        https.get(`https://youx.live/alias/${alias.alias}?breakCache=true`, res => {
                            process.stdout.write("Done\n");
                        });
                    });
                }
            });
        }
    });
}

function newGame(owner: string) {
    return getObjectAsync<string>(`admin/new/${owner}`);
}

function getToken(id: string) {
    return getStringAsync(`admin/token/${id}`);
}

function getAliasToken(id: string) {
    return getStringAsync(`admin/aliastoken/${id}`);
}

function modifyJson(fileName: string, action: (doc: any) => void) {
    fs.readJson(fileName, (err, doc) => {
        if (err) {
            throw err;
        }

        action(doc);

        fs.writeJson(fileName, doc, { spaces: 4 }, err => {
            if (err) {
                throw err;
            }
        });
    });
}

export async function buildGame(targetPath: string, name: string) {
    const id = await newGame("320a5291-ed91-4e6e-b77d-543d4d671be6");
    console.log(`Creating new game with id ${id}`);
    const gamePath = path.join(targetPath, name);

    fs.access(gamePath, err => {
        if (!err) {
            throw "The template already exists";
        } else {
            fs.copySync(TEMPLATES_PATH, gamePath);

            const indexFileName = getIndexFileName(gamePath);
            modifyJson(indexFileName, index => {
                index.id = id;
                index.name = name;
            });

            const launchFileName = getLaunchFileName(gamePath);
            modifyJson(launchFileName, launch => {
                launch.configurations[0].url = `https://youchoose.live/${id}`;
            });
        }
    });
}

function getInstance(gamePath: string, instance: string) {
    const instanceFileName = getInstanceFileName(gamePath, instance);
    if (!fs.existsSync(instanceFileName)) {
        throw `Cannot find instance ${instance}`;
    }
    const details = fs.readJsonSync(instanceFileName);
    return details;
}

function getGame(gamePath: string) {
    const indexFileName = getIndexFileName(gamePath);
    const index = fs.readJsonSync(indexFileName);
    return index;
}

function buildSymbols(game: any, name: string, symbolFile: string, srcDir: string, files: string[]) {
    const sprites = svgstore({
        svgAttrs: { display: "none"}
    });
    files.forEach(file => sprites.add(file, fs.readFileSync(path.join(srcDir, file), "utf8")));
    fs.writeFileSync(path.join(srcDir, symbolFile), sprites.toString({inline: true}));
    game[name] = symbolFile;
}


function buildSvgSymbols(game: any, srcDir: string) {
    if (game.logo) {
        buildSymbols(game, "gameSymbols", "gameSymbols.svg", srcDir, [ game.logo ]);
    }
    game.rounds.forEach((round: any) => {
        const svgFiles: string[] = [];
        function addSvgFileIfSet(file?: string) {
            if (file && file.toLowerCase().endsWith(".svg")) {
                svgFiles.push(file);
            }
        }
        function addQuestions(questions?: any[]) {
            questions && questions.forEach(question => {
                addSvgFileIfSet(question.image);
                addQuestions(question.questions);
            });
        }

        addSvgFileIfSet(round.logo);
        addQuestions(round.questions);
        if (svgFiles.length) {
            buildSymbols(round, "roundSymbols", `${round.name}Symbols.svg`, srcDir, svgFiles);
        }
    });
}

export async function uploadGame(gamePath: string, instance?: string) {
    let instanceId: string | undefined;
    if (instance) {
        const instanceFileName = getInstanceFileName(gamePath, instance);
        if (fs.existsSync(instanceFileName)) {
            const details = fs.readJsonSync(instanceFileName);
            instanceId = details.id;
        } else {
            const id = await newGame("320a5291-ed91-4e6e-b77d-543d4d671be6");
            const details = {
                name: instance,
                id: id
            };
            instanceId = id;
            fs.writeJsonSync(instanceFileName, details);
        }
    }

    const game = getGame(gamePath);
    const gameId = instanceId || game.id;
    console.log(`gameId = ${gameId}`);
    game.id = gameId;

    const md5Map: { [key: string]: string } = {};
    const token = await getToken(gameId);
    const blobService = azureStorage.createBlobServiceWithSas("skeventresources.blob.core.windows.net", token);

    const srcDir = path.join(gamePath, "src");
    const files: klaw.Item[] = [];
    buildSvgSymbols(game, srcDir);

    const uploadGameFiles = (callback: () => void) => {
        if (files.length) {
            const item = files.pop()!;
            const file = item.path;
            if (file === "index.json") {
                setTimeout(() => uploadGameFiles(callback), 0);
            } else {
                const stats = item.stats;
                const relative = file.substring(srcDir.length + 1);
                const blobName = relative.replace("\\", "/");
                blobService.getBlobProperties(gameId, blobName, (error, result, response) => {
                    if (error && (response.statusCode !== 404)) {
                        console.log(`Error getting metadata for ${blobName}`, response);
                        throw error;
                    }
                    if (result && result.contentSettings && result.contentSettings.contentMD5) {
                        md5Map[blobName] = result.contentSettings.contentMD5;
                    }

                    if (!result || (new Date(result.lastModified) < stats.mtime)) {
                        blobService.createBlockBlobFromLocalFile(gameId, blobName, file,
                            (error, result, response) => {
                                if (error) {
                                    console.log(`Error creating blog ${blobName}`, response);
                                    throw error;
                                } else {
                                    process.stdout.write(`${blobName}.`);
                                    uploadGameFiles(callback);
                                }
                            });
                    } else {
                        process.stdout.write(`.`);
                        setTimeout(() => uploadGameFiles(callback), 0);
                    }
                });
            }
        } else {
            console.log("Looking at game");
            const addMD5 = (o: any, name: string) => {
                const value = o[name];
                if (value) {
                    const cache = md5Map[value];
                    if (cache) {
                        const b64 = Buffer.from(cache).toString("base64");
                        o[name + "MD5"] = b64;
                    }
                }
            };

            addMD5(game, "logo");
            addMD5(game, "gameSymbols");
            game.rounds && game.rounds.forEach((round: any) => {
                addMD5(round, "logo");
                addMD5(round, "data");
                addMD5(round, "backgroundImage");
                addMD5(round, "roundSymbols");
                function addQuestionMD5(questions?: any[]) {
                    questions && questions.forEach((question: any) => {
                        addMD5(question, "image");
                        addQuestionMD5(question.questions);
                    });
                }
                addQuestionMD5(round.questions);
            });

            const json = JSON.stringify(game);
            fs.writeJsonSync(".uploaded.Json", game);
            const blobName = "index.json";
            blobService.createBlockBlobFromText(gameId, blobName, json,
                (error, result, response) => {
                    if (error) {
                        console.log(`Error creating blog ${blobName}`, response);
                        throw error;
                    } else {
                        console.log("uploaded index.json");
                        callback();
                    }
                });
        }
    };


    klaw(srcDir)
        .on("data", i => {
            if (i.stats.isFile()) {
                files.push(i);
            }
        })
        .on("end", () => {
            uploadGameFiles(() => {
                process.stdout.write("Completed\n");
                bustConfigurationCacheAsync(gameId)
                    .then(() => bustAliasCacheAsync(gamePath, instance));
            });
        });

    process.stdout.write("Searching for changes..");
}


export async function publishGame(gamePath: string, alias: string, instance?: string, urls?: string[]) {
    const instanceDetails = instance && getInstance(gamePath, instance);

    const game = getGame(gamePath);
    const id = instance && instanceDetails.id || game.id;
    const name = game.name;
    const owner = game.owner;

    const fullAlias = {
        alias: alias,
        id: id,
        name: name,
        owner: owner,
        urls: urls
    };

    const json = JSON.stringify(fullAlias, undefined, 4);
    const aliasFile = getAliasFileName(gamePath, instance);
    console.log("Writing alias file: ", aliasFile);
    fs.writeFileSync(aliasFile, json);

    console.log(`Creating alias ${alias}`);
    const token = await getAliasToken(id);
    const blobService = azureStorage.createBlobServiceWithSas("skeventresources.blob.core.windows.net", token);
    blobService.createBlockBlobFromText("aliases", `${alias}.json`, json, (error, result, response) => {
        if (error) {
            console.log(response);
            throw error;
        }

        https.get(`${BASE_URL}/admin/refreshall`, res => {
            if (res.statusCode !== 200) {
                console.log(`Error: ${res.statusCode} (${res.statusMessage})`, res.headers);
                res
                    .on("error", err => { throw err; });
            } else {
                console.log("Published");
            }
        });
    });
}

